;NMP wrapper (c) NOP 2020 - Maarten Loor
;Requires DOS2.xx and at least 16KB free memory segment

;Use: DEMO.COM MODFILE.MOD

;This example runs IM 2 and does not exit
;If you want to use FD9A you need to make sure that hook maps slots and memory

	org #100

	; / ** EQU ** /
bdos:		equ	#5
extbio:		equ	#ffca

ALL_SEG:		equ	#0	;Allocate a 16k segment.
FRE_SEG:		equ	#3	;Free a 16k segment.
RD_SEG:		equ	#6	;Read byte from address A:HL to A.
WR_SEG:		equ	#9	;Write byte from E to address A:HL.
CAL_SEG:		equ	#C	;Inter-segment call.  Address in IYh:IX
CALLS:		equ	#F	;Inter-segment call.  Address in line after the call instruction.
PUT_PH:		equ	#12	;Put segment into page (HL).
GET_PH:		equ	#15	;Get current segment for page (HL)
PUT_P0:		equ	#18	;Put segment into page 0.
GET_P0:		equ	#1B	;Get current segment for page 0.
PUT_P1:		equ	#1E	;Put segment into page 1.
GET_P1:		equ	#21	;Get current segment for page 1.
PUT_P2:		equ	#24	;Put segment into page 2.
GET_P2:		equ	#27	;Get current segment for page 2.
PUT_P3:		equ	#2A	;Not supported since page-3 must never be changed.
GET_P3:		equ	#2D	;Get current segment for page 3.

q_vdp98:		equ	#98
q_vdp99:		equ	#99

	; Interrupt location
q_interruptTable:	equ #c000
q_interruptAddress:	equ #c1c1

q_memoryMax:	equ	8	;(384K max)

	jp	start

	include "nmp.asm"
	include "sub_callbacks.asm"

start:
	;remember stack pointer for correct exit when loading fails
	ld	(v_SP),sp

	;request memory
	call	f_memoryInit
	or	a
	ret	nz

	;open the mod file
	ld	de,#82
	ld	bc,#43	;_OPEN
	ld	a,%001	;read only
	call	bdos
	or	a
	jp	nz,f_explainError
	ld	hl,v_msgError
	jp	nz,f_doTermWithError
	ld	a,b
	ld	(v_fileHandleMOD),a

	ld	de,d_msgHeaderLoad
	call	f_printString

	;load module
	call	mt_load

	;close mod file
	ld	c,#45
	ld	a,(v_fileHandleMOD)
	ld	b,a
	call	bdos
	or	a
	jp	nz,f_explainError
	ld	hl,v_msgError
	jp	nz,f_doTermWithError

	ld	de,d_msgExit
	call	f_printString

	;start playing
	im	2
	call	mt_start
loop:
	;here you would scan your keys and maybe exit
	ei
	halt
	jp	loop

	;stop playing
	call	mt_stop
	im	1
	ret

f_exit:
	di
	ld	sp,(v_SP)
	push	af
	push	hl
	;disable interrupt
	im	1
	
	;restore	memory
	call	f_memoryTerm

	pop	hl
	pop	af
	ret	z
	push	hl
	ld	de,d_msgNewLine
	call	f_printString
	ld	de,d_msgError
	call	f_printString
	pop	de
	call	f_printString
	ld	de,d_msgNewLine
	jp	f_printString

f_memoryInit:	PROC
;Initialises the memory
;In:
; A=number of 16KB pages
;Returns:
; Z=success
; NZ=failed
;Changes:
; ?

	;set interrupt
	ld	hl,q_interruptTable
	ld	de,q_interruptTable+1
	ld	bc,#100
	ld	(hl),q_interruptAddress	&	#FF
	ldir

	ld	hl,interrupt
	ld	(q_interruptAddress+1),hl
	ld	a,#c3
	ld	(q_interruptAddress),a

	ld	a,q_interruptTable	/	#100
	ld	i,a

	ld	c,06fh
	call	bdos		; Get MSX-DOS version
	ld	hl,d_Error_InvalidDos
	and	a
	jp	nz,f_doTermWithError	; No MSX-DOS
	ld	a,b
	cp	2
	jp	c,f_doTermWithError

	;reset the temp page as so f_printString won't use it yet
	xor	a
	ld	(v_mmuInitialPage1),a

	;get mapper variable table
	xor	a
	ld	de,#0401
	call	extbio
	ld	(v_mmuVariables),hl
	
	;get mapper support routine address
	xor	a
	ld	de,#0402
	call	extbio
	ld	(v_mmuFunctions),hl

;	ld	a,c
;	cp	q_memoryMax
;	ld	a,#DE	;ERROR: NOT ENOUGH MEMORY
;	jp	c,f_explainError

	;current pages
	ld	hl,GET_P0
	call	f_mapper
	ld	(v_mmuInitialPage0),a
	ld	hl,GET_P1
	call	f_mapper
	ld	(v_mmuInitialPage1),a
	ld	hl,GET_P2
	call	f_mapper
	ld	(v_mmuInitialPage2),a

	ld	de,d_msgAllocatingMemory
	call	f_printString

	ld	hl,v_mmuPages
	ld	b,q_memoryMax
loop:	
	push	bc
	push	hl
	;allocate memory segment
	xor	a	;allocate user segment
	ld	b,a	;primary mapper
	ld	hl,ALL_SEG
	call	f_mapper
	pop	hl
	pop	bc
	jr	c,loopexit
	ld	(hl),a
	inc	hl
	djnz	loop
loopexit:
	ld	a,q_memoryMax
	sub	b
	ld	(v_mmuMax),a

	ld	a,(v_mmuMax)
	cp	1
	jp	c,error

	;set first free memory postion
	ld	hl,#0000
	ld	(v_ADRMEM),hl
	xor	a
	ld	(v_PAGMEM),a

	in	a,(#c4)
	inc	a
	jp	nz,done

	;no OPL4 found
	ld	de,d_Error_OplNotFound
	call	f_printString
	ld	de,d_msgNewLine
	call	f_printString
	ld	a,255
	ret
done:
	xor	a
	ret
error:
	ld	a,#DE	;ERROR: NOT ENOUGH MEMORY
	jp	c,f_explainError
	ENDP

f_memoryTerm:	PROC
;Restore memory configuration
	im	1
	call	f_setDataPage
	jp	f_setTempPage
	ENDP

f_explainError:	PROC
;Explains error code from f_bdos
;In:
; A=Error code
;Out:
; -
;Changes:
; DE (maybe more) 
	push	af
;	ld	de,d_msgError
;	call	f_printString
	ld	c,#66
	pop	af
	push	af
	ld	b,a
	ld	de,v_msgError
	call	bdos
;	ld	de,v_msgError
;	call	f_printString
;	ld	de,d_msgNewLine
;	call	f_printString
	pop	af
;	ld	hl,d_Error_FileHandling
	ld	hl,v_msgError
	jp	f_doTermWithError
	ENDP

f_doTermWithError:	PROC
	di
	push	hl
	call	mt_stop
	im	1
	pop	hl
direct:
	ld	a,255	;return error
	or	a
	jp	f_exit	;@@@@ EXIT WITH ERROR FIX
	ENDP

f_printString:	PROC
;Print string to screen (zero terminated)
;In:
; DE=location of string
;Out:
; -
;Changes:
; AF,DE
loop:
	ld	a,(de)
	or	a
	ret	z
	call	f_printChar
	inc	de
	jr	loop
	ENDP

f_printChar:	PROC
;Outputs a character
;In:
; A=ASCII code
;Out:
; -
;Changes:
; AF,IX
	push	bc
	push	de
	push	hl
	ld	e,a
	ld	c,#2
	call	bdos
	pop	hl
	pop	de
	pop	bc
	ret
	ENDP

f_setDataPage:	PROC
;Set memory page1 4000-7FFF to the DATA page
	push	af
	push	hl
	;set data buffer at page1
	ld	a,(v_mmuInitialPage1)
	ld	hl,PUT_P1
	call	f_mapper
	pop	hl
	pop	af
	ret
	ENDP

f_setTempPage:	PROC
;Set memory page1 8000-BFFF to the TEMP page
	push	af
	push	hl
	;set data buffer at page2
	ld	a,(v_mmuInitialPage2)
	ld	hl,PUT_P2
	call	f_mapper
	pop	hl
	pop	af
	ret
	ENDP

f_selectPages:	PROC
;Select memory. DE is changed to the new location
;In:
; [CF]DE=offset (17 bits)
;Out:
; DE=memory location (16 bits)
;Changes:
; AF,C,DE
	push	hl
	ld	a,d
	rra	;shift C back in A
	and	%11100000
	rlca
;	and	%11000000
	rlca
	rlca
	ld	c,a
	ld	a,(mt_PAGMEM_MOD)
	add	a,c
;	ld (v_PAGMEM),a
	;<begin>call f_selectPages
	push	de
	ld	hl,v_mmuPages
	ld	e,a
	ld	d,0
	add	hl,de
;	ld	a,(hl)
;	push	hl
;	ld	hl,PUT_P1
;	call	f_mapper
;	pop	hl
;	inc	hl
	ld	a,(hl)
	ld	hl,PUT_P2
	call	f_mapper
	pop	de
	;<end>call f_selectPages
	set	7,d	;offset #8000
	res	6,d
	ld	hl,(mt_ADRMEM_MOD)
	add	hl,de
	ex	de,hl
	pop	hl
	ret
	ENDP

f_mapper:	PROC
;In:
; HL=Call number
;Changes:
; Nothing
	push de
	ld de,(v_mmuFunctions)
	add hl,de
	ld (address+1),hl
	pop de
address:
	jp	#fd9f
	ENDP

;##########	interrupt

interrupt:	PROC
	push	af
	push	bc
	push	de
	push	hl
	exx
	ex	af,af'
	push	af
	push	bc
	push	de
	push	hl
	push	ix
	push	iy

	;OPL interrupt? gets priority
	in	a,(#c4)
	and	%10000000
	jp	nz,music	;YES

	;VDP interrupt
	in	a,(q_vdp99)
	and	%10000000
	jp	z,exit	;NO

	;vdp interrupt stuff
	;and reading hid devices

	;maybe also OPL interrupt?
	in	a,(#c4)
	and	%10000000
	jp	z,exit	;NO
music:

	call	mt_play
exit:
	pop	iy
	pop	ix
	pop	hl
	pop	de
	pop	bc
	pop	af
	ex	af,af'
	exx
	pop	hl
	pop	de
	pop	bc
	pop	af
	ei
	ret
	ENDP

;########## STRINGS

d_msgError:		db	"ERROR:",0
d_msgHeaderLoad:	db	"Loading",13,10,0
d_msgPatternLoad:	db	"Loading patterns",13,10,0
d_msgSampleLoad:	db	"Loading sample",13,10,0
;d_msgPlaying:	db	"Playing...",0
d_msgAllocatingMemory:	db	"Allocating memory",13,10,0
d_msgExit:		db	"Exit",13,10,0
d_msgNewLine:	db	13,10,0
d_Error_FileHandling:	db	"File handling",13,10,0
d_Error_InvalidDos:	db	"Unsupported DOS version",13,10,0
d_Error_OplNotFound:	db	"OPL4 cartridge not found",13,10,0
d_Error_NoFilesFound:	db	"No files found",13,10,0
d_Error_OutOfMemory:	db	"Out of memory",13,10,0

;########## RAM

v_fileHandleMOD:	db	0
v_ADRMEM:		dw	0	;free memory address
v_PAGMEM:		db	0	;free memory page (index for v_mmuPages)
v_PAGMEM_Max:	db	0

v_SP:		dw	0	;stack pointer for error
v_mmuVariables:	dw	0
v_mmuFunctions:	dw	0
v_mmuInitialPage0:	db	0	;initial page0
v_mmuInitialPage1:	db	0	;initial page1
v_mmuInitialPage2:	db	0	;initial page2
v_mmuPages:		ds	q_memoryMax
v_mmuMax:		db	0	;number of pages
v_hook:		db	5

v_msgError:				;buffer to hold the error (OVERWRITES THE FOLLOWING)
p_modheader:	ds	1084