;N.O.P. Mod player adapted by Maarten Loor from the Amiga 68000 version created by Peter Hanning
;Build 2021.12.05.1

;todo
;- the notenumber is always searched (iy+q_mt_notenum), als when the sample does not have a ft
;- cannot start other samples with period (when it is set to loop)

;changelog:
;2021.12.05.1:	added mt_song_pos_queue (queues a position but only starts to play at normal postion change)

;mt_init:		Initialize OPL
;mt_load:		load a module
;mt_play:		interrupt player
;mt_stop:		stop playing

;timer2 N2=256-(7812,5/bpm)  -> 8*n2=2048-(62500/bmp)

opl4_reg:		equ	#7e
opl4_data:		equ	#7f

q_opl4_timer_125:	equ	62	;timer1=249, timer2=62
q_opl4_timer_reg:	equ	3	;timer1=2, timer2=3
q_opl4_timer_on:	equ	%01000010	;timer1=%00100001, timer2=%01000010

;### data

;p_modheader:	equ #4000
;p_samplebuffer:	equ #a000
;q_samploadbuffersize:	equ #2000

q_mt_temp_size:	equ	16

q_mt_sampnum:	equ	0	;B    SampleNumber   
q_mt_notenum:	equ	1	;B    NoteNumber   
q_mt_period:	equ	2	; W   Period   
q_mt_volume:	equ	4	;B    Volume   
q_mt_oldsamp:	equ	5	;B    OldSampleNumber   
q_mt_cmdtemp:	equ	6	; W   CommandTemp   
q_mt_oldpbendsp:	equ	8	;B    OldPitchBendSpeed
q_mt_oldvibcmd:	equ	9	;B    OldVibratoCommand
q_mt_vibtabpos:	equ	10	;B    VibratoTablePosition
q_mt_oldtremcmd:	equ	11	;B    OldTremoloCommand
q_mt_tremtabpos:	equ	12	;B    TremoloTablePosition
q_mt_wavectrl:	equ	13	;B    WaveControl %0000TTVV
q_mt_pattloop:	equ	14	; W   PatternLoop(E6)temp

mt_pos_temp_ch0:	ds	q_mt_temp_size,0
mt_pos_temp_ch1:	ds	q_mt_temp_size,0
mt_pos_temp_ch2:	ds	q_mt_temp_size,0
mt_pos_temp_ch3:	ds	q_mt_temp_size,0

mt_modheader_adr:	dw	0	;mod header address
mt_numpatterns:	db	0	;number of mt_patterns loaded      /loading/
mt_sampnum:		db	0	;temp sampnum
mt_sramPointer:	db	#20,#06,#00	;temp address in chipmem  
mt_samplen:		dw	0	;temp length of sample  
mt_samprpp:		dw	0	;temp sample repeat point  
mt_samprpl:		dw	0	;temp sample repeat length  

mt_finetunes:	ds	31,0	;finetunes for samples [0-30]   /playing/
mt_volumes:		ds	31,0	;volumes of samples [0-30]  
mt_patterns:	ds	128	;pattern data [0-127]
mt_pansetting:	db	11,5,5,11
mt_song_length:	db	0	;number of patterns (songlength)  
mt_song_pos:	db	0	;current position [cur,copy]  
mt_song_pospat:	db	0
mt_song_step:	db	0	;current step [cur,copy]  
mt_channel:		db	255	;current mt_channel (255 if not playing)
mt_vblanks:		db	6,6	;vblank timing (cur, org)  
mt_tempo:		dw	0,q_opl4_timer_125,125	;cia timing    (cur, org,ami) (*80uS)  
mt_arpeggio:	db	0	;for playing c64 sounds  
mt_command_E5h:	db	0	;temp. for finetune  
mt_command_EEh:	db	0	;temp. voor patt.delay  
mt_stopflag:	db	0

mt_OAT_Pointer:	dw	0
mt_ADRMEM_MOD:	dw	0	;location of the pattern data
mt_PAGMEM_MOD:	db	0

mt_cmd9enable:	db	0	;0=disable; player supports command 9 (sample offset)
mt_cmd9temp:	ds	194	;command temp table (128-31 samples, 2 bytes per sample/offset combination)
mt_repeat_song:	db	1
mt_paused:		db	0
mt_vol:		db	0
mt_song_pos_queue:	db	255	;if set; the next position will be overrules by this queue

opl4_mod_init:
		db	2,%00010000
		db	#f8,%111111	;fm mix control
		db	#f9,%000000	;wav mix contol
		db	0

opl4_wave_hdr:
		db	0,0,0	;0: start         (3b) 
		db	0,0	;3: loop          (2b) 
		db	0,0	;5: end           (2b) 
		db	0	;lfo,vib       (1b) 
		db	%11111111	;ar, d1r       (1b) 
		db	0	;dl, d2r       (1b) 
		db	0	;rate comp, rr (1b) 
		db	0	;am            (1b) 

mt_periods:
		dw	0	;for identification if no note specified 
; Tuning 0, Normal
		dw	856,808,762,720,678,640,604,570,538,508,480,453
		dw	428,404,381,360,339,320,302,285,269,254,240,226
		dw	214,202,190,180,170,160,151,143,135,127,120,113
mt_ft_tab:
; Tuning 1
		db	-6 ,-6 ,-5 ,-5 ,-4 ,-3 ,-3 ,-3 ,-3 ,-3 ,-3 ,-3
		db	-3 ,-3 ,-2 ,-3 ,-2 ,-2 ,-2 ,-1 ,-1 ,-1 ,-1 ,-1
		db	-1 ,-1 ,-1 ,-1 ,-1 ,-1 ,-1 ,-1 ,-1 ,-1 ,-1 ,0
; Tuning 2
		db	-12,-12,-10,-11,-8 ,-8 ,-7 ,-7 ,-6 ,-6 ,-6 ,-6
		db	-6 ,-6 ,-5 ,-5 ,-4 ,-4 ,-4 ,-3 ,-3 ,-3 ,-3 ,-2
		db	-3 ,-3 ,-2 ,-3 ,-3 ,-2 ,-2 ,-2 ,-2 ,-2 ,-2 ,-1
; Tuning 3
		db	-18,-17,-16,-16,-13,-12,-12,-11,-10,-10,-10,-9
		db	-9 ,-9 ,-8 ,-8 ,-7 ,-6 ,-6 ,-5 ,-5 ,-5 ,-5 ,-4
		db	-5 ,-4 ,-3 ,-4 ,-4 ,-3 ,-3 ,-3 ,-3 ,-2 ,-2 ,-2
; Tuning 4
		db	-24,-23,-21,-21,-18,-17,-16,-15,-14,-13,-13,-12
		db	-12,-12,-11,-10,-9 ,-8 ,-8 ,-7 ,-7 ,-7 ,-7 ,-6
		db	-6 ,-6 ,-5 ,-5 ,-5 ,-4 ,-4 ,-4 ,-4 ,-3 ,-3 ,-3
; Tuning 5
		db	-30,-29,-26,-26,-23,-21,-20,-19,-18,-17,-17,-16
		db	-15,-14,-13,-13,-11,-11,-10,-9 ,-9 ,-9 ,-8 ,-7
		db	-8 ,-7 ,-6 ,-6 ,-6 ,-5 ,-5 ,-5 ,-5 ,-4 ,-4 ,-4
; Tuning 6
		db	-36,-34,-32,-31,-27,-26,-24,-23,-22,-21,-20,-19
		db	-18,-17,-16,-15,-14,-13,-12,-11,-11,-10,-10,-9
		db	-9 ,-9 ,-7 ,-8 ,-7 ,-6 ,-6 ,-6 ,-6 ,-5 ,-5 ,-4
; Tuning 7
		db	-42,-40,-37,-36,-32,-30,-29,-27,-25,-24,-23,-22
		db	-21,-20,-18,-18,-16,-15,-14,-13,-13,-12,-12,-10
		db	-10,-10,-9 ,-9 ,-9 ,-8 ,-7 ,-7 ,-7 ,-6 ,-6 ,-5
; Tuning -8
		db	51,48,46,42,42,38,36,34,32,30,28,27
		db	25,24,23,21,21,19,18,17,16,15,14,14
		db	12,12,12,10,10,10,9 ,8 ,8 ,8 ,7 ,7
; Tuning -7
		db	44,42,40,37,37,35,32,31,29,27,25,24
		db	22,21,20,19,18,17,16,15,15,14,13,12
		db	11,10,10,9 ,9 ,9 ,8 ,7 ,7 ,7 ,6 ,6
; Tuning -6
		db	38,36,34,32,31,30,28,27,25,24,22,21
		db	19,18,17,16,16,15,14,13,13,12,11,11
		db	9 ,9 ,9 ,8 ,7 ,7 ,7 ,6 ,6 ,6 ,5 ,5
; Tuning -5
		db	31,30,29,26,26,25,24,22,21,20,18,17
		db	16,15,14,13,13,12,12,11,11,10,9 ,9
		db	8 ,7 ,8 ,7 ,6 ,6 ,6 ,5 ,5 ,5 ,5 ,5
; Tuning -4
		db	25,24,23,21,21,20,19,18,17,16,14,14
		db	13,12,11,10,11,10,10,9 ,9 ,8 ,7 ,7
		db	6 ,6 ,6 ,5 ,5 ,5 ,5 ,4 ,4 ,4 ,3 ,4
; Tuning -3
		db	19,18,17,16,16,15,15,14,13,12,11,10
		db	9 ,9 ,9 ,9 ,8 ,8 ,7 ,7 ,7 ,6 ,5 ,6
		db	5 ,4 ,5 ,4 ,4 ,4 ,4 ,3 ,3 ,3 ,3 ,3
; Tuning -2
		db	12,12,12,10,11,11,10,10,9 ,8 ,7 ,7
		db	6 ,6 ,6 ,5 ,6 ,5 ,5 ,5 ,5 ,4 ,4 ,4
		db	3 ,3 ,3 ,3 ,2 ,3 ,3 ,2 ,2 ,2 ,2 ,2
; Tuning -1
		db	6 ,6 ,6 ,5 ,6 ,6 ,6 ,5 ,5 ,5 ,4 ,4
		db	3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,2 ,2 ,2
		db	2 ,1 ,2 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1

mt_pitch_tab:
		;per 2 bytes, eerste 2 bytes voor period 907   
		;laatste voor period 108 (dus ook voor alle fine-tunes!)   
;  C-1 to B-1 : 856,808,762,720,678,640,604,570,538,508,480,453
;  C-2 to B-2 : 428,404,381,360,339,320,302,285,269,254,240,226
;  C-3 to B-3 : 214,202,190,180,170,160,151,143,135,127,120,113

		include	"pitchtab.asm"

mt_vol_tab:
		db	#7F,#5C,#52,#4A,#44,#3F,#3B,#37,#34,#31
		db	#2F,#2D,#2A,#28,#27,#25,#23,#22,#20,#1F
		db	#1E,#1C,#1B,#1A,#19,#18,#17,#16,#15,#14
		db	#13,#12,#12,#11,#10,#0F,#0F,#0E,#0D,#0C
		db	#0C,#0B,#0B,#0A,#09,#09,#08,#08,#07,#06
		db	#06,#05,#05,#04,#04,#03,#03,#03,#02,#02
		db	#01,#01,#00,#00,#00

mt_vib_tab:
		db	000,024,049,074,097,120,141,161
		db	180,197,212,224,235,244,250,253
		db	255,253,250,244,235,224,212,197
		db	180,161,141,120,097,074,049,024

;### subroutines
mt_mul16:		ld	a,16
mt_mul16_1:		sla	l
		rl	h
		sla	e
		rl	d
		jp	nc,mt_mul16_2
		add	hl,bc
mt_mul16_2:		dec	a
		jp	nz,mt_mul16_1
		ret

mt_div16:
		ld	bc,-1
mt_div16_1:
		inc	bc
		scf
		ccf
		sbc	hl,de
		jp	nc,mt_div16_1
		add	hl,de
		ret

mt_rst20:
		ld	a,h
		sub	d
		ret	nz
		ld	a,l
		sub	e
		ret

opl4_dampall:
;damp all mt_channels
		ld	b,24	;#mt_channels
opl4_dampall_lp:
		;damp
		opl4_wt
		ld	a,#68-1
		add	a,b
		out	(opl4_reg),a
		opl4_wt
		ld	a,%11111111
		out	(opl4_data),a

		;volume
		opl4_wt
		ld	a,#50-1
		add	a,b
		out	(opl4_reg),a
		opl4_wt
		ld	a,254
		out	(opl4_data),a

		;attack rate
		opl4_wt
		ld	a,#98-1
		add	a,b
		out	(opl4_reg),a
		opl4_wt
		ld	a,%00000000
		out	(opl4_data),a
		
		;rate correction
		opl4_wt
		ld	a,#c8-1
		add	a,b
		out	(opl4_reg),a
		opl4_wt
		ld	a,%00000000
		out	(opl4_data),a

		;decay
		opl4_wt
		ld	a,#b0-1
		add	a,b
		out	(opl4_reg),a
		opl4_wt
		ld	a,%00000000
		out	(opl4_data),a

		djnz	opl4_dampall_lp

;:		;mute mixer
;:		opl4_wt
;:		ld	a,#f9
;:		out	(opl4_reg),a
;:		opl4_wt
;:		ld	a,%11111111
;:		out	(opl4_data),a
;:		ei
		ret

opl4_wt:		MACRO
opl4_wtlp:		in	a,(#c4)
;:		rra
;:		jp	c,opl4_wtlp
		and	%11
		jp	nz,opl4_wtlp
		ENDM

opl4_initWave:
		opl4_wt
		ld	a,5
		out	(#c6),a
		opl4_wt
		ld	a,3
		out	(#c5),a
		ret

opl4_setRegisters:
		opl4_wt
		ld	a,(hl)
		or	a
		ret	z
		out	(opl4_reg),a
		inc	hl
		opl4_wt
		ld	a,(hl)
		out	(opl4_data),a
		inc	hl
		jp	opl4_setRegisters

opl4_setAddress:
		push	af
		opl4_wt
		ld	a,3
		out	(opl4_reg),a
		opl4_wt
		pop	af
		out	(opl4_data),a
		opl4_wt
		ld	a,4
		out	(opl4_reg),a
		opl4_wt
		ld	a,h
		out	(opl4_data),a
		opl4_wt
		ld	a,5
		out	(opl4_reg),a
		opl4_wt
		ld	a,l
		out	(opl4_data),a
		ret

opl4_setVolume:
		opl4_wt
		ld	a,#f9
		out	(opl4_reg),a
		ld	a,(mt_vol)
		rlca
		rlca
		rlca
		ld	b,a
		opl4_wt
		ld	a,(mt_vol)
		or	b
		out	(opl4_data),a
		ret

opl4_ramPointerAdd:
		inc	de
		inc	de
		ld	c,0	;nextbyteextra =0 
		ld	a,(de)	;get destin d7-d0 
		add	a,l	;add source d7-d0 
		ld	(de),a	;put destin d7-d0 
		call	c,opl4_ramPointerAdd1	;if>255 then inc nextbyteextra 
		dec	de

		ld	a,(de)	;get destin d15-d8 
		add	a,c	;add lastbyteextra
		ld	c,0	;nextbyteextra=0 
		call	c,opl4_ramPointerAdd1	;if>255 then inc nextbyteextra 
		add	a,h	;add source d15-d8
		call	c,opl4_ramPointerAdd1
		ld	(de),a
		dec	de

		ld	a,(de)
		add	a,c
		ld	c,0
		call	c,opl4_ramPointerAdd1
		add	a,b
		call	c,opl4_ramPointerAdd1
		ld	(de),a
		ret

opl4_ramPointerAdd1:
		inc	c
		ret

opl4_ramPointerSub:
		ld	a,b
		cpl
		ld	b,a
		ld	a,h
		cpl
		ld	h,a
		ld	a,l
		cpl
		ld	l,a
		jp	opl4_ramPointerAdd

; Fill SRAM with FF
opl4_clearRam:
		opl4_wt
		ld	a,6
		out	(opl4_reg),a
		ld	c,opl4_data
		ld	b,#20
opl4_clearRamlp2:
		ld	de,#8000
opl4_clearRamlp:
		opl4_wt
		ld	a,#80
		out	(c),a
		dec	de
		opl4_wt
		ld	a,#80
		out	(c),a
		ld	a,d
		or	e
		jp	nz,opl4_clearRamlp
		djnz	opl4_clearRamlp2
		ret

; Move data from RAM to Sample-RAM
; In HL = source
;     DE = length
; Modi AF, C, DE, HL

opl4_ramToSram:
		opl4_wt
		ld	a,6	; RS Eenmt_mul16 is genoeg!
		out	(opl4_reg),a	; RS (tenminste, zonder interrupts)
		ld	c,opl4_data
opl4_ramToSramlp:
		opl4_wt
		outi
		dec	de
		ld	a,d
		or	e
		jp	nz,opl4_ramToSramlp
		ret

; Move data from Sample-RAM to RAM
; In HL = destination
;     DE = length
; Modi AF, C, DE, HL

opl4_sramToRam:
		opl4_wt
		ld	a,6
		out	(opl4_reg),a
		ld	c,opl4_data
opl4_sramToRamlp:
		opl4_wt
		ini
		dec	de
		ld	a,d
		or	e
		jp	nz,opl4_sramToRamlp
		ret

		;mt_loadSample 
		;------- 
		;[hl]=length of sample
		;set destination chipmem address before 
mt_loadSample:
		ld	a,h
		or	l
		ret	z	;return when sample size=0

		push	hl	;push length of sample
		;sample < #3001?
		ld	de,q_samploadbuffersize+1
		call	mt_rst20
		jp	c,mt_loadSample1
		ld	hl,q_samploadbuffersize	;load a block of #3000
mt_loadSample1:
		push	hl	;push bytes to load

		;load data
		push	hl
		ld	de,p_samplebuffer
		call	mt_loadData
		pop	hl

		;tranfer the sample data to chipmem
		ex	de,hl ;set number of bytes loaded to DE
		ld	hl,p_samplebuffer
		di
		call	opl4_ramToSram
		ei
		pop	de	;pop bytes to load
		pop	hl	;pop length of sample
		or	a
		sbc	hl,de
		jp	mt_loadSample

mt_scanOffset:
;Scan all patterns for offset command
;Replace offset command with new sample number
;In:
; Nothing
;Out:
; Nothing
;Changes:
; All registers
		ld	a,(mt_cmd9enable)
		or	a
		ret	z
		;clear tmp table
		ld	hl,mt_cmd9temp
		ld	de,mt_cmd9temp+1
		ld	bc,193	;((128-31)*2)-1
		ld	(hl),0
		ldir

		;remember the current sram address
		ld	a,(mt_sramPointer)
		ld	hl,(mt_sramPointer+1)
		push	af
		push	hl
		;set start address at #20147 (sample 32)
		ld	hl,#4701
		ld	(mt_sramPointer+1),hl
		ld	a,#20
		ld	(mt_sramPointer),a
		
		ld	a,(mt_numpatterns)
		ld	b,a
mt_scanOffsetLoop:
		push	bc
		ld	a,(mt_numpatterns)
		sub	b
		rlca	;512
		rlca	;1024
		ld	e,0
		ld	d,a
		call	mt_memorySet
		push	de
		pop	ix	;IX=start address
		ld	b,64	;64 steps

mt_scanOffsetPatternLoop:
		push	bc
		ld	b,4	;4 mt_channels
		ld	iy,mt_pos_temp_ch0
mt_scanOffsetChannelLoop:
		push	bc
		call	mt_scanOffsetNote
		ld	de,4	;each note is 4 bytes
		add	ix,de
		ld	de,q_mt_temp_size
		add	iy,de
		pop	bc
		djnz	mt_scanOffsetChannelLoop
		
		pop	bc
		djnz	mt_scanOffsetPatternLoop

		pop	bc
		djnz	mt_scanOffsetLoop
		;restore the current sram address
		pop	hl
		pop	af
		ld	(mt_sramPointer),a
		ld	(mt_sramPointer+1),hl
		ret

mt_scanOffsetNote:
;Process a single note for the offset command
;In:
; IX=note address
;Out:
; IX=note address
;Changes:
; AF,BC,DE,HL
		ld	a,(ix+2)
		and	%1111
		cp	9
		ret	nz	;no cmd 9 found

		ld	a,(ix+3)
		or	a
		ret	z	;parameter is zero

		;compose sampleid
		ld	a,(ix+2)	;get sampleid
		rrca
		rrca
		rrca
		rrca
		and	%00001111
		ld	e,a
		ld	a,(ix+0)
		and	%11110000
		or	e
		ld	e,a
		or	a
		jp	z,mt_scanOffsetNoteSkipSampleSet
		ld	(iy+q_mt_sampnum),e
mt_scanOffsetNoteSkipSampleSet:
		ld	a,e
		or	a
		ret	z	;sample is still 0
		ld	e,(iy+q_mt_sampnum)
		ld	d,(ix+3)	;org sampleoffset

		call	mt_scanOffsetGetTableIndex
		ld	(ix+3),0
		ret	nz	;table full
		or	a
		jp	z,mt_scanOffsetNoteNew
mt_scanOffsetNoteFound:
		;calculate new sampleid
		ld	de,mt_cmd9temp
		or	a
		sbc	hl,de
		srl	h	;/2
		rr	l
		ld	de,#20	;add #20 to the sample number
		add	hl,de
		;replace the offset command
		ld	(ix+3),l
		ret
mt_scanOffsetNoteNew:
		;store the orginal note info in the temp table
		ld	e,(iy+q_mt_sampnum)
		ld	(hl),e
		inc	hl
		ld	(hl),d
		push	de

		;find the orginal header
		dec	e
		ld	d,0
		ld	bc,12
		call	mt_mul16	;*12
		ld	a,#20
		call	opl4_setAddress

		;read back to ram
		ld	hl,opl4_wave_hdr	;copy back to ram
		ld	de,12
		call	opl4_sramToRam

		pop	de
		;modify the header
		;start
		;if offset>length: set offset begin of loop
		ld	a,(opl4_wave_hdr+5)	;get end
		cpl
		ld	e,a	;E=end
		ld	a,d	;offset
		cp	e	;offset-end
		jp	c,mt_scanOffsetNoteNewStart	;fits
		;check if offset>sample length: start at sample loop
		ld	a,(opl4_wave_hdr+3)	;repeat
		ld	d,a	;D=repeat
		ld	a,e	;A=end
		sub	d	;A=end-repeat
		ld	d,a
mt_scanOffsetNoteNewStart:
		ld	a,(opl4_wave_hdr+1)	;start
		add	a,d		;add offset
		ld	(opl4_wave_hdr+1),a	;set new start
		jr	nc,mt_scanOffsetNoteNewEnd
		ld	a,(opl4_wave_hdr)
		inc	a
		ld	(opl4_wave_hdr),a

mt_scanOffsetNoteNewEnd:
		;end
		ld	a,(opl4_wave_hdr+5)
		cpl
		sub	d
		cpl
		ld	(opl4_wave_hdr+5),a

		;repeat
		ld	a,(opl4_wave_hdr+3)
		sub	d
		ld	(opl4_wave_hdr+3),a

		;write back to new position
		ld	a,(mt_sampnum)
		ld	e,a
		ld	d,0
		ld	bc,12
		call	mt_mul16	;*12
		ld	a,#20
		call	opl4_setAddress

		;write to sram
		ld	hl,opl4_wave_hdr	;copy back to ram
		ld	de,12
		call	opl4_ramToSram

		;replace the new sample number as parameter in the 9 command
		ld	a,(mt_sampnum)
		inc	a
		ld	(ix+3),a
		ld	(mt_sampnum),a
		ret

mt_scanOffsetGetTableIndex:
;Find DE in cmd9temp table
;In:
; Nothing
;Out:
; HL=position in table
; A=0: Not found,create new entry
; Z=found
; NZ=table full
		ld	hl,mt_cmd9temp
		ld	b,128-31	;max new samples that can be defined
mt_scanOffsetGetTableIndexLoop:
		ld	a,(hl)
		or	a
		ret	z
		ld	a,(hl)
		cp	e
		jp	nz,mt_scanOffsetGetTableIndexNext
		inc	hl
		ld	a,(hl)
		dec	hl
		cp	d
		jp	nz,mt_scanOffsetGetTableIndexNext
		;found
		ret	z
mt_scanOffsetGetTableIndexNext:
		inc	hl
		inc	hl
		djnz	mt_scanOffsetGetTableIndexLoop
		;not found & table full
		ld	a,255
		or	a
		ret

;### main prg

mt_set_modheader_adr:
;Set the mod header address
;HL=address
		ld	(mt_modheader_adr),hl
		ret

mt_load:
;Loads a MOD
;In:
; Nothing
		call	mt_memoryInit
		call	mt_init
		call	opl4_initWave

		xor	a
		ld	(mt_paused),a
		ld	a,(mt_channel)
		cp	255
		call	nz,mt_stop
		call	opl4_dampall	;damp the 24 first OPL4 mt_channels
		call	opl4_setVolume

		opl4_wt
		ld	a,2	;set opl4 for memory access
		out	(opl4_reg),a
		opl4_wt
		ld	a,%00000001
		out	(opl4_data),a

;debug;clear sram
;:		ld	a,#20
;:		ld	hl,#0000
;:		call	opl4_setAddress
;:		call	opl4_clearRam

		;set start address at #20600
		ld	hl,#0006
		ld	(mt_sramPointer+1),hl
		ld	a,#20
		ld	(mt_sramPointer),a

		;uncompress object
		ld	de,(mt_modheader_adr)
		ld	hl,1084
		call	mt_loadData

		;get song length (number of patterns to play)
		ld	ix,(mt_modheader_adr)
		ld	de,950
		add	ix,de
		ld	a,(ix)
		ld	(mt_song_length),a

		;copy the pattern index table (max 128 bytes)
		ld	hl,(mt_modheader_adr)
		ld	de,952
		add	hl,de
		ld	de,mt_patterns
		ld	bc,128
		ldir

		;search the pattern table to find the highest pattern
		;defines how much pattern data to load
		ld	ix,(mt_modheader_adr)	;find highest pattern number
		ld	de,952
		add	ix,de
		ld	c,0
		ld	b,128
mt_find_pat:
		ld	a,(ix)
		inc	ix
		cp	c
		jp	c,mt_find_pat1
		ld	c,a
mt_find_pat1:
		djnz	mt_find_pat

		inc	c
		ld	a,c
		ld	(mt_numpatterns),a

mt_loadPatterns:
		ld	b,0	;B=current pattern
				;C=number of patterns
mt_loadPatternsLoop:
		ld	a,c	;number of patterns
		sub	b	;subtract current pattern
		ld	d,a
		cp	17	;if <17
		jp	c,mt_loadPatternsStart
		ld	d,16	;load 16 patterns
mt_loadPatternsStart:
		;load D patterns
		push	bc
		push	de

		ld	a,d
		ld	l,0	;HL=number of bytes to load (size)
		ld	h,a
		sla	h	;*512
		sla	h	;*1024

		ld	e,0	;CF,DE=destination address
		ld	d,b
		sla	d	;*512
		sla	d	;*1024

		call	mt_loadPattern

		pop	de
		pop	bc
		
		ld	a,b	;add loaded patterns D to the pattern counter B
		add	a,d
		ld	b,a
		cp	c	;are we done loading?
		jp	nz,mt_loadPatternsLoop

		;EXAMINE sampleinfo
		;------------------

		ld	ix,(mt_modheader_adr)	;mod sample attribute data
		ld	de,20
		add	ix,de
		xor	a
		ld	(mt_sampnum),a
		ld	b,31	;31 samples!
mt_makeHeaders:
		push	bc	;sampnum counter

		ld	h,(ix+42-20)
		ld	l,(ix+43-20)
		add	hl,hl

		ld	(mt_samplen),hl	;set sample length

		ld	hl,mt_finetunes
		ld	a,(mt_sampnum)
		ld	e,a
		ld	d,0
		add	hl,de

		ld	a,(ix+44-20)	;sample finetune [-7...+7]
		ld	(hl),a

		ld	b,(ix+45-20)	;sample volume [0...64]

		ld	hl,mt_volumes
		ld	a,(mt_sampnum)
		ld	e,a
		ld	d,0
		add	hl,de
		ld	(hl),b

		ld	h,(ix+46-20)
		ld	l,(ix+47-20)
		add	hl,hl
		ld	(mt_samprpp),hl	;sample repeat point

		ld	h,(ix+48-20)	;sample repeat length
		ld	l,(ix+49-20)
		add	hl,hl
		ld	(mt_samprpl),hl

		ld	hl,mt_sramPointer	;set wave header start address
		ld	de,opl4_wave_hdr+0
		ld	bc,3
		ldir

		push	ix	;load sample in chipmem
		call	mt_progressLoadSample
		ld	a,(mt_sramPointer+1)
		ld	h,a
		ld	a,(mt_sramPointer+2)
		ld	l,a
		ld	a,(mt_sramPointer+0)
		call	opl4_setAddress	;set the opl4 sram destination address
		ld	hl,(mt_samplen)
		call	mt_loadSample	;load the sample data and transfer it
		pop	ix

		ld	hl,(mt_samplen)
		ld	a,h
		or	l
		jp	z,mt_lengthready
		dec	hl
		ld	(mt_samplen),hl

		ld	hl,(mt_samprpp)	;if sample repeat =0
		ld	a,h
		or	l
		jr	nz,mt_sampleSetRepeat
		ld	hl,(mt_samprpl)	;if sample replen<(2*2)
		ld	de,4
		call	mt_rst20
		jr	nc,mt_sampleSetRepeat
;sample is not repeating
mt_sampleSetNonRepeat:
		;loop address=sample length
		ld	hl,(mt_samplen)	;loopaddress=pt2.length
		inc	hl
		ld	(mt_samplen),hl
		jr	mt_lengthready
;sample is repeating
mt_sampleSetRepeat:
		ld	hl,(mt_samprpp)	;adres
		ld	de,(mt_samprpl)	;number of bytes
		add	hl,de
		dec	hl
		ld	(mt_samplen),hl	;address
		ld	hl,(mt_samprpp)	;loopaddress=pt2.repeat
mt_lengthready:
		ld	a,h	;set wave header loop address
		ld	(opl4_wave_hdr+3),a
		ld	a,l
		ld	(opl4_wave_hdr+4),a

		ld	hl,(mt_samplen)	;set wave header end address
		ld	a,l
		cpl
		ld	(opl4_wave_hdr+6),a
		ld	a,h
		cpl
		ld	(opl4_wave_hdr+5),a

		ld	a,(mt_sampnum)	;calculate the destination chipmem addres for the note header
		ld	c,a
		ld	b,0
		ld	de,12
		call	mt_mul16
		ld	a,#20
		call	opl4_setAddress

		ld	hl,opl4_wave_hdr	;put header in OPL4 memory
		ld	de,12
		call	opl4_ramToSram

		ld	b,0	;add length
		ld	hl,(mt_samplen)
		ld	de,mt_sramPointer
		call	opl4_ramPointerAdd
		ld	hl,(mt_samplen)
		ld	a,h
		or	l
		jp	z,mt_nosample

		call	mt_fixSample

mt_nosample:
		ld	de,30	;add 30 bytes to samplepointer
		add	ix,de
		ld	a,(mt_sampnum)
		inc	a
		ld	(mt_sampnum),a

		pop	bc	;pop sampnum teller
		dec	b	;next sample
		ld	a,b
		or	a
		jp	nz,mt_makeHeaders

		call	mt_scanOffset

		xor	a
		ld	(mt_song_pos),a
		ld	(mt_song_step),a
		ret

mt_fixSample:
		;fix end of non repeating samples
		ld	hl,1	;go back a byte
		ld	b,0
		ld	de,mt_sramPointer
		call	opl4_ramPointerSub

		;repeat the last byte times
		ld	a,(mt_sramPointer+1)
		ld	h,a
		ld	a,(mt_sramPointer+2)
		ld	l,a
		ld	a,(mt_sramPointer+0)
		call	opl4_setAddress
		opl4_wt
		ld	a,6	;read the last byte from sample memory
		out	(opl4_reg),a
		ld	c,opl4_data
		opl4_wt
		in	b,(c)
		opl4_wt
		out	(c),b	;repeat it 3 times
		opl4_wt
		out	(c),b
		opl4_wt
		out	(c),b

		ld	b,0	;add 4 bytes to the sample memory pointer
		ld	hl,4
		ld	de,mt_sramPointer
		jp	opl4_ramPointerAdd

;### playing routine

mt_start:
		ld	a,(mt_channel)
		cp	255
		jp	nz,mt_start1
		call	mt_init
mt_start1:
		ld	hl,mt_pos_temp_ch0
		ld	de,mt_pos_temp_ch0+1
		ld	bc,q_mt_temp_size*4-1
		ld	(hl),0
		ldir
		xor	a
		ld	(mt_stopflag),a

		ld	a,(mt_paused)
		or	a
		jp	nz,mt_start2

		xor	a
		ld	(mt_song_pos),a
		;reset the queue
		ld	a,255
		ld	(mt_song_pos_queue),a
		ld	a,128
		ld	(mt_song_step),a
		ld	hl,0
		ld	(mt_tempo+0),hl
		ld	hl,q_opl4_timer_125
		ld	(mt_tempo+2),hl
		ld	hl,125
		ld	(mt_tempo+4),hl
		ld	a,6
		ld	(mt_vblanks+0),a
		ld	(mt_vblanks+1),a
mt_start2:
		;timer
		di
		opl4_wt
		ld	a,4	;R4
		out	(#c4),a
		opl4_wt
		ld	a,128	;RST
		out	(#c5),a

		opl4_wt
		ld	a,q_opl4_timer_reg
		out	(#c4),a
		opl4_wt
		ld	a,(mt_tempo+2)	;tempo=125
		neg
		out	(#c5),a

		opl4_wt
		ld	a,4
		out	(#c4),a
		opl4_wt
		ld	a,q_opl4_timer_on	;enable TIMER1
		out	(#c5),a
		ei
		xor	a
		ld	(mt_paused),a
		ld	(mt_channel),a
		ret

mt_stop:
		di
		;reset timer
		opl4_wt
		ld	a,4	;R4
		out	(#c4),a
		opl4_wt
		ld	a,%01100000	;TIMER1&TIMER2
		out	(#c5),a

		ld	hl,opl4_mod_init
		call	opl4_setRegisters

		call	opl4_dampall	;damp the 24 first OPL4 mt_channels
		;mute mixer
		opl4_wt
		ld	a,#f9
		out	(opl4_reg),a
		opl4_wt
		ld	a,%11111111
		out	(opl4_data),a
;:		call	opl4_setVolume
		ei

		ld	a,255
		ld	(mt_channel),a

		ld	hl,0
		ld	(mt_tempo),hl
		ret

mt_play:
mt_music:
		ld	hl,(mt_tempo)
		ld	a,h
		or	l
		jp	z,mt_tempozero
		jp	mt_settempous
mt_tempozero:
		ld	hl,(mt_tempo+2)
		ld	(mt_tempo),hl
		call	mt_settempous

		ld	a,(mt_vblanks)
		dec	a
		ld	(mt_vblanks),a
		or	a
		jp	z,mt_music_l0

mt_offnote:
		ld	a,(mt_arpeggio)
		inc	a
		cp	3
		jp	nz,mt_arpeggio_max
		xor	a
mt_arpeggio_max:
		ld	(mt_arpeggio),a

		;maybe store old page settings
		call	mt_address	;will change the page

		xor	a
		ld	(mt_channel),a

		ld	iy,mt_pos_temp_ch0

		ld	b,4
mt_offnote1:
		push	bc
		ld	a,(ix+2)
		and	%1111
		jp	z,mt_offncom_0
		cp	1
		jp	z,mt_offncom_1
		cp	2
		jp	z,mt_offncom_2
		cp	3
		jp	z,mt_offncom_3
		cp	4
		jp	z,mt_offncom_4
		cp	5
		jp	z,mt_offncom_5
		cp	6
		jp	z,mt_offncom_6
		cp	7
		jp	z,mt_offncom_7
		cp	#A
		jp	z,mt_offncom_A
		cp	#B
		jp	z,mt_offncom_B
		cp	#D
		jp	z,mt_offncom_D
		cp	#E
		jp	nz,mt_of_return
		ld	a,(ix+3)
		and	%11110000
		cp	#60
		jp	z,mt_offncom_E6
		cp	#90
		jp	z,mt_offncom_E9
		cp	#C0
		jp	z,mt_offncom_EC
		cp	#D0
		jp	z,mt_offncom_ED
mt_of_return:
		pop	bc
		call	mt_addadr
		ld	a,(mt_channel)
		cp	255
		jp	z,mt_of_return1
		inc	a
		ld	(mt_channel),a
mt_of_return1:
		djnz	mt_offnote1
		ret	;jp	mt_exit

mt_offncom_0:
		ld	a,(ix+3)
		or	a
		jp	z,mt_of_return

		ld	a,(mt_arpeggio)
		ld	l,(iy+q_mt_period)
		ld	h,(iy+q_mt_period+1)
		or	a
		jp	z,mt_offncom_09
		ld	l,(iy+q_mt_notenum)
		ld	h,0
		cp	1
		ld	a,(ix+3)
		jp	nz,mt_offncom_02
		rrca
		rrca
		rrca
		rrca
mt_offncom_02:
		and	%1111
mt_offncom_08:
		ld	e,a
		ld	d,0
		add	hl,de
		sla	l
		rl	h
		ld	de,mt_periods
		add	hl,de
		ld	e,(hl)
		inc	hl
		ld	d,(hl)
		ex	de,hl
		ld	a,(iy+q_mt_notenum)
		push	af
		call	mt_setnotep	;add finetunes
		pop	af
		ld	(iy+q_mt_notenum),a
mt_offncom_09:
		call	opl4_fnum
		jp	mt_of_return

mt_offncom_1:
		ld	e,(ix+3)
		ld	d,0
		ld	l,(iy+q_mt_period)
		ld	h,(iy+q_mt_period+1)
		scf
		ccf
		sbc	hl,de
		ld	de,108
		call	mt_rst20
		jp	nc,mt_offncom_11
		ld	hl,108
mt_offncom_11:
		ld	(iy+q_mt_period),l
		ld	(iy+q_mt_period+1),h
		call	opl4_fnum
		jp	mt_of_return

mt_offncom_2:
		ld	e,(ix+3)
		ld	d,0
		ld	l,(iy+q_mt_period)
		ld	h,(iy+q_mt_period+1)
		add	hl,de
		ld	de,906+1
		call	mt_rst20
		jp	c,mt_offncom_21
		ld	hl,906
mt_offncom_21:
		ld	(iy+q_mt_period),l
		ld	(iy+q_mt_period+1),h
		call	opl4_fnum
		jp	mt_of_return

mt_offncom_3:
		call	mt_offncom_30
		jp	mt_of_return
mt_offncom_30:
		ld	l,(iy+q_mt_period)	;get current note
		ld	h,(iy+q_mt_period+1)
		ld	e,(iy+q_mt_cmdtemp)	;get new note
		ld	d,(iy+q_mt_cmdtemp+1)
		call	mt_rst20
		jp	z,opl4_fnum	;curnote = newnote

		jp	c,mt_offncom_31

		push	de	;curnote > newnote (pitch down)
		ld	e,(iy+q_mt_oldpbendsp)
		ld	d,0
		scf
		ccf
		sbc	hl,de
		pop	de
		jp	c,mt_offncom_34
		call	mt_rst20
		jp	nc,mt_offncom_32
mt_offncom_34:
		ex	de,hl
		jp	mt_offncom_32
mt_offncom_31:
		push	de	;curnote < newnote  (pitch up)
		ld	e,(iy+q_mt_oldpbendsp)
		ld	d,0
		add	hl,de
		pop	de
		inc	de
		call	mt_rst20
		jp	c,mt_offncom_32
		dec	de
		ex	de,hl
mt_offncom_32:
		ld	(iy+q_mt_period),l
		ld	(iy+q_mt_period+1),h
		jp	opl4_fnum

mt_offncom_4:
		call	mt_offncom_40
		jp	mt_of_return
mt_offncom_40:
		ld	a,(iy+q_mt_vibtabpos)
		rrca		;lsr.w #2,{a}
		rrca
		and	#1f	;and.w $001F,{a}
		ld	c,a
		ld	b,0
		ld	a,(iy+q_mt_wavectrl)
		and	%11
		jp	z,mt_offncom_42

		sla	c	;lsl.b #3,{bc}
		sla	c
		sla	c
		cp	1
		jp	z,mt_offncom_41	;vib_rampdown
		ld	c,255
		jp	mt_offncom_43
mt_offncom_41:
		;mt_vib_rampdown
		ld	a,(iy+q_mt_vibtabpos)
		bit	7,a
		jp	z,mt_offncom_43
		ld	a,255
		sub	c
		ld	c,a
		jp	mt_offncom_43
mt_offncom_42:
		;mt_vib_sine
		ld	hl,mt_vib_tab
		add	hl,bc
		ld	c,(hl)
mt_offncom_43:
		;mt_vib_set
		ld	a,(iy+q_mt_oldvibcmd)
		and	%1111	;and.w #15,{a}
		ld	b,0
		ld	e,a
		ld	d,0
		call	mt_mul16	;mulu {bc},{de} -> {hl}
		srl	h
		rr	l
		srl	h
		rr	l
		srl	h
		rr	l
		srl	h
		rr	l
		srl	h
		rr	l
		srl	h
		rr	l
		srl	h
		rr	l	;lsr.w #7,{hl}
		ex	de,hl
		ld	l,(iy+q_mt_period)	;move.w q_mt_period,{hl}
		ld	h,(iy+q_mt_period+1)

		ld	a,(iy+q_mt_vibtabpos)
		bit	7,a
		jp	nz,mt_offncom_44
		add	hl,de
		jp	mt_offncom_45
mt_offncom_44:
		;mt_VibratoNeg
		scf
		ccf
		sbc	hl,de
mt_offncom_45:
		;mt_vibrato3
		ld	(iy+q_mt_cmdtemp),l
		ld	(iy+q_mt_cmdtemp+1),h
		call	opl4_fnum
		ld	a,(iy+q_mt_oldvibcmd)
		rrca
		rrca
		and	%111100
		add	a,(iy+q_mt_vibtabpos)
		ld	(iy+q_mt_vibtabpos),a
		ret

mt_offncom_5:
		call	mt_offncom_30
		jp	mt_offncom_A

mt_offncom_6:
		call	mt_offncom_40
		jp	mt_offncom_A

mt_offncom_7:
		ld	a,(iy+q_mt_tremtabpos)
		rrca		;lsr.w #2,{a} 
		rrca
		and	#1f	;and.w $001F,{a} 
		ld	c,a
		ld	b,0
		ld	a,(iy+q_mt_wavectrl)
		rrca
		rrca
		and	%11
		jp	z,mt_offncom_72

		sla	c	;lsl.b #3,{bc} 
		sla	c
		sla	c
		cp	1
		jp	z,mt_offncom_71	;trem_rampdown
		ld	c,255
		jp	mt_offncom_73
mt_offncom_71:
		;mt_trem_rampdown
		ld	a,(iy+q_mt_tremtabpos)
		bit	7,a
		jp	z,mt_offncom_73
		ld	a,255
		sub	c
		ld	c,a
		jp	mt_offncom_73
mt_offncom_72:
		;mt_trem_sine
		ld	hl,mt_vib_tab
		add	hl,bc
		ld	c,(hl)
mt_offncom_73:
		;mt_trem_set
		ld	a,(iy+q_mt_oldtremcmd)
		and	%1111	;and.w #15,{a} 
		ld	b,0
		ld	e,a
		ld	d,0
		call	mt_mul16	;mulu {bc},{de} -> {hl} 
		srl	h
		rr	l
		srl	h
		rr	l
		srl	h
		rr	l
		srl	h
		rr	l
		srl	h
		rr	l
		srl	h
		rr	l	;lsr.w #6,{hl}
		ex	de,hl
		ld	l,(iy+q_mt_volume)	;move.w q_mt_volume,{hl}
		ld	h,0

		ld	a,(iy+q_mt_tremtabpos)
		bit	7,a
		jp	nz,mt_offncom_74
		add	hl,de
		jp	mt_offncom_75
mt_offncom_74:
		;mt_tremoloNeg
		scf
		ccf
		sbc	hl,de
		jp	nc,mt_offncom_75
		ld	hl,0
mt_offncom_75:
		;mt_tremolo3
		ld	a,l
		cp	#41
		jp	c,mt_offncom_76
		ld	l,#40
mt_offncom_76:
		ld	e,l
		call	opl4_totallev0

		ld	a,(iy+q_mt_oldtremcmd)
		rrca
		rrca
		and	%111100
		add	a,(iy+q_mt_tremtabpos)
		ld	(iy+q_mt_tremtabpos),a
		jp	mt_of_return

mt_offncom_A:
		ld	a,(ix+3)
		cp	#10
		jp	nc,mt_offncom_A1
		and	%1111
		ld	b,a
		ld	a,(iy+q_mt_volume)
		sub	b
		jp	nc,mt_offncom_A2
		xor	a
		jp	mt_offncom_A2
mt_offncom_A1:
		rrca
		rrca
		rrca
		rrca
		and	%1111
		ld	b,a
		ld	a,(iy+q_mt_volume)
		add	a,b
		cp	#41
		jp	c,mt_offncom_A2
		ld	a,#40
mt_offncom_A2:
		ld	(iy+q_mt_volume),a
		call	opl4_totallev
		jp	mt_of_return

mt_offncom_B:
		ld	a,(mt_song_pos)
		ld	b,a
		ld	a,(mt_song_length)
		dec	a
		cp	b
		jp	nz,mt_offncom_B1
		ld	a,(mt_repeat_song)
		or	a
		jp	nz,mt_offncom_B1
		call	mt_stop
		jp	mt_of_return
mt_offncom_B1:
		ld	a,(mt_vblanks)
		cp	1
		jp	nz,mt_of_return
		ld	b,(ix+3)
		;handle position queue
		ld	a,(mt_song_pos_queue)
		cp	255
		jp	z,mt_offncom_B2
		ld	b,a
mt_offncom_B2:
		ld	a,b
		ld	(mt_song_pos),a
		;reset queue
		ld	a,255
		ld	(mt_song_pos_queue),a
		ld	a,128
		ld	(mt_song_step),a
		jp	mt_of_return

mt_offncom_D:
		ld	a,(mt_song_pos)
		ld	b,a
		ld	a,(mt_song_length)
		dec	a
		cp	b
		jp	nz,mt_offncom_D2
		ld	a,(mt_repeat_song)
		or	a
		jp	nz,mt_offncom_D2
		call	mt_stop
		jp	mt_of_return
mt_offncom_D2:
		ld	a,(mt_vblanks)
		cp	1
		jp	nz,mt_of_return
		ld	a,(mt_song_step)
		bit	7,a
		jp	nz,mt_of_return
		ld	a,(ix+3)
		set	7,a
		ld	(mt_song_step),a
		ld	a,(mt_song_length)
		ld	b,a
		ld	a,(mt_song_pos)
		inc	a
		cp	b
		jp	c,mt_offncom_D1
		xor	a
mt_offncom_D1:
		;handle position queue
		ld	b,a
		ld	a,(mt_song_pos_queue)
		cp	255
		jp	z,mt_offncom_D3
		ld	b,a
mt_offncom_D3:
		ld	a,b
		ld	(mt_song_pos),a
		;reset queue
		ld	a,255
		ld	(mt_song_pos_queue),a
		jp	mt_of_return

mt_offncom_E6:
		ld	a,(mt_vblanks)
		cp	1
		jp	nz,mt_of_return
		ld	a,(ix+3)
		and	%1111
		ld	b,a
		jp	nz,mt_offncom_E61
		ld	a,(mt_song_step)
		ld	(iy+q_mt_pattloop),a
		jp	mt_of_return
mt_offncom_E61:
		ld	a,(iy+q_mt_pattloop+1)
		inc	a
		inc	b
		cp	b
		jp	nz,mt_offncom_E62
		xor	a
		ld	(iy+q_mt_pattloop+1),a
		jp	mt_of_return
mt_offncom_E62:
		ld	(iy+q_mt_pattloop+1),a
		ld	a,(iy+q_mt_pattloop)
		set	7,a
		ld	(mt_song_step),a
		jp	mt_of_return

mt_offncom_E9:
		ld	a,(iy+q_mt_cmdtemp)
		or	a
		jp	z,mt_of_return
		dec	a
		ld	(iy+q_mt_cmdtemp),a
		or	a
		jp	nz,mt_of_return
mt_offncom_E91:
		ld	a,(ix+3)
		and	%1111
		ld	(iy+q_mt_cmdtemp),a

		call	opl4_key
		ex	af,af'
		opl4_wt
		ex	af,af'
		res	7,a
		out	(opl4_data),a

		call	opl4_key
		ex	af,af'
		opl4_wt
		ex	af,af'
		set	7,a
		out	(opl4_data),a
		jp	mt_of_return

mt_offncom_EC:
		ld	a,(iy+q_mt_cmdtemp)
		or	a
		jp	z,mt_of_return
		dec	a
		ld	(iy+q_mt_cmdtemp),a
		or	a
		jp	nz,mt_of_return
		xor	a
		jp	mt_offncom_A2

mt_offncom_ED:
		ld	a,(iy+q_mt_cmdtemp)
		or	a
		jp	z,mt_of_return
		dec	a
		ld	(iy+q_mt_cmdtemp),a
		or	a
		jp	nz,mt_of_return
		call	mt_newnote
		jp	mt_of_return

mt_music_l0:
		ld	a,(mt_vblanks+1)
		ld	(mt_vblanks),a

		ld	a,(mt_command_EEh)
		or	a
		jp	z,mt_music_l1
		dec	a
		ld	(mt_command_EEh),a
		ret	;jp	mt_exit

mt_music_l1:
		call	mt_addstep
		ld	a,(mt_channel)
		cp	255
		ret	z	;jp	z,mt_exit
		call	mt_address
		xor	a
		ld	(mt_channel),a

		ld	iy,mt_pos_temp_ch0

		ld	b,4
mt_music_l2:
		push	bc
		ld	l,(iy+q_mt_period)
		ld	h,(iy+q_mt_period+1)
		call	opl4_fnum
		call	opl4_totallev
		xor	a
		ld	(mt_command_E5h),a
		call	mt_setsampl
		ld	a,(ix+2)
		and	%1111
		jp	z,mt_command_0
		cp	4
		jp	z,mt_command_4
		cp	7
		jp	z,mt_command_7
		cp	8
		jp	z,mt_command_8
		cp	9
		jp	z,mt_command_9
		cp	#B
		jp	z,mt_command_B
		cp	#C
		jp	z,mt_command_C
		cp	#D
		jp	z,mt_command_D
		cp	#f
		jp	z,mt_command_F
		cp	#E
		jp	nz,mt_return
		ld	a,(ix+3)
		and	%11110000
		cp	#40
		jp	z,mt_command_E4
		cp	#50
		jp	z,mt_command_E5
		cp	#70
		jp	z,mt_command_E7
		cp	#80
		jp	z,mt_command_E8
		cp	#90
		jp	z,mt_command_E9
		cp	#A0
		jp	z,mt_command_EA
		cp	#B0
		jp	z,mt_command_EB
		cp	#C0
		jp	z,mt_command_EC
		cp	#D0
		jp	z,mt_command_ED
		cp	#E0
		jp	z,mt_command_EE
mt_return:
		call	mt_newnote
		ld	a,(ix+2)
		and	%1111
		cp	#e
		jp	nz,mt_return_ED
		ld	a,(ix+3)
		and	%11110000
		cp	#10
		call	z,mt_command_E1
		cp	#20
		call	z,mt_command_E2
mt_return_ED:
		call	opl4_key
		set	7,a
		ex	af,af'
		opl4_wt
		ex	af,af'
		out	(opl4_data),a
		call	opl4_totallev
		call	mt_addadr

		ld	a,(mt_channel)
		inc	a
		ld	(mt_channel),a

		ld	a,(mt_stopflag)
		or	a
		call	nz,mt_stop
		pop	bc
		dec	b
		ld	a,b
		or	a
		jp	nz,mt_music_l2
		ret	;mt_exit

mt_addstep:
		ld	a,(mt_song_step)
		bit	7,a
		jp	z,mt_addstep0
		res	7,a
		ld	(mt_song_step),a
		ret
mt_addstep0:
		inc	a
		cp	64
		jp	nz,mt_addstep2

		ld	a,(mt_song_length)
		ld	b,a
		ld	a,(mt_song_pos)
		inc	a
		cp	b
		jp	c,mt_addstep1
		ld	a,(mt_repeat_song)
		or	a
		jp	nz,mt_addstep3
		call	mt_stop
		ret
mt_addstep3:
		xor	a
mt_addstep1:
		;handle position queue
		ld	b,a
		ld	a,(mt_song_pos_queue)
		cp	255
		jp	z,mt_addstep4
		ld	b,a
mt_addstep4:
		ld	a,b
		ld	(mt_song_pos),a
		;reset the queue
		ld	a,255
		ld	(mt_song_pos_queue),a
		xor	a
mt_addstep2:
		ld	(mt_song_step),a
		ret

mt_init:
		ld	hl,opl4_mod_init
		call	opl4_setRegisters
		ret

mt_address:
		ld	a,(mt_song_step)
		and	%01111111
		ld	l,a
		ld	h,0
		add	hl,hl	;*2
		add	hl,hl	;*4
		add	hl,hl	;*8
		add	hl,hl	;*16

		push	hl
		ld	a,(mt_song_pos)
		ld	e,a
		ld	d,0
		ld	hl,mt_patterns
		add	hl,de
		ld	a,(hl)
		ld	(mt_song_pospat),a
		ld	h,a
		ld	l,0
		sla	h	;*512
		sla	h	;*1024
		ex	af,af'	;push CF
		pop	de
		add	hl,de
		ex	af,af'	;pop CF
		ex	de,hl
		call	mt_memorySet

		push	de	;ld	ix,de
		pop	ix
		ret

mt_addadr:
		ld	de,4
		add	ix,de
		ld	de,q_mt_temp_size
		add	iy,de
		ret

mt_setsampl:
		ld	a,(ix+2)	;get sampleid
		rrca
		rrca
		rrca
		rrca
		and	%00001111
		ld	b,a
		ld	a,(ix+0)
		and	%11110000
		or	b
		or	a
		ret	z
		ld	(iy+q_mt_sampnum),a
		dec	a
		ld	e,a
		ld	d,0
		ld	hl,mt_volumes
		add	hl,de
		ld	a,(hl)
		ld	(iy+q_mt_volume),a	;set standard volume
		ret

mt_newnote:
		ld	a,(iy+q_mt_sampnum)	;get previous set samplenumber
		or	a
		ret	z	;exit if no samplenumber is set 

		ld	a,(ix+0)
		and	%00001111
		ld	h,a
		ld	l,(ix+1)
		or	l
		ret	z	;exit if no period specified 

		ld	a,(ix+2)
		and	%1111
		cp	3
		jp	z,mt_command_3
		cp	5
		jp	nz,mt_no_comm_3
mt_command_3:
		ld	a,(ix+3)
		or	a
		jp	z,mt_command_5
		ld	(iy+q_mt_oldpbendsp),a
mt_command_5:
		call	mt_setnotep	;add finetunes  
		ld	a,h
		or	l
		ret	z
		ld	(iy+q_mt_cmdtemp),l	;in command temp.
		ld	(iy+q_mt_cmdtemp+1),h
		ret

mt_no_comm_3:
		push	hl
		ld	e,0
		call	opl4_totallev0
		call	opl4_key
		res	7,a
		ex	af,af'
		opl4_wt
		ex	af,af'
		out	(opl4_data),a

		ld	hl,mt_channel

		opl4_wt
		ld	a,#8	;set sample number
		add	a,(hl)
		out	(opl4_reg),a
		ld	a,(iy+q_mt_oldsamp)
		ld	b,a
		ld	a,(iy+q_mt_sampnum)
		cp	b
		jp	z,mt_newnote1
		ld	(iy+q_mt_oldsamp),a
		add	a,127
		ex	af,af'
		opl4_wt
		ex	af,af'
		out	(opl4_data),a
mt_newnote1:
		pop	hl
		call	mt_setnotep	;convert [hl] (nontuned) into tuned 
		ld	(iy+q_mt_period),l
		ld	(iy+q_mt_period+1),h
		call	opl4_fnum
		ret

mt_setnotep:
		ld	a,h
		or	l
		ret	z	;exit if no period in mt_pos_temp_ch
		push	hl	;push period nr. in tabel 
		ld	b,h
		ld	c,l

		call	mt_findnote	;zoek period in tabel 
		ld	(iy+q_mt_notenum),l
		ld	b,h
		ld	c,l

		ld	e,(iy+q_mt_sampnum)	;get samplenumber
		ld	a,(mt_cmd9enable)
		or	a
		jp	z,mt_setnotep2
		;cmd9
		ld	a,(ix+2)
		and	%1111
		cp	9
		jp	nz,mt_setnotep2
		ld	a,(ix+2)
		rrca
		rrca
		rrca
		rrca
		and	%1111
		ld	e,a
		ld	a,(ix+0)
		and	%11110000
		or	e
		ld	e,a
mt_setnotep2:
		;cmd9end
		dec	e
		ld	d,0
		ld	hl,mt_finetunes	;get finetune of sample 
		add	hl,de
		ld	de,0
		ld	a,(mt_command_E5h)
		or	a
		jp	nz,mt_setnotep0
		ld	a,(hl)
mt_setnotep0:
		or	a
		jp	z,mt_setnotep1	;fintune=0 -> do not alter 

		dec	a
		push	bc	;push noteoffset

		ld	h,0	; HL = A * 36
		sla	a
		sla	a
		ld	l,a
		sla	a
		sla	a
		sla	a
		rl	h
		add	a,l
		ld	l,a
		ld	a,h
		adc	a,0
		ld	h,a

		ld	de,mt_ft_tab
		add	hl,de
		pop	de	;pop noteoffset
		add	hl,de	;add offset
		ld	a,(hl)
		ld	e,a
		ld	d,0

		bit	7,a
		jp	z,mt_setnotep1
		ld	d,255
mt_setnotep1:
		pop	hl
		add	hl,de
		ret

mt_findnote:
		ld	hl,mt_periods	;find period in tabel; [hl]=notenr.

		ld	e,c
		ld	d,b

		ld	bc,36*256+255

mt_findnote1:
		ld	a,e
		cpi
		jp	nz,mt_findnote2
		ld	a,d
		cpi
		jp	nz,mt_findnote3
		ld	de,-mt_periods-2
		add	hl,de
		srl	h
		rr	l
		ret

mt_findnote2:
		inc	hl
mt_findnote3:
		djnz	mt_findnote1
		ld	hl,0
		ret

opl4_fnum:
		ld	de,-107	;[HL] = amiga period
		add	hl,de

		sla	l
		rl	h
		ld	de,mt_pitch_tab
		add	hl,de
		ld	d,(hl)
		inc	hl
		ld	e,(hl)

		ld	hl,mt_channel

		opl4_wt
		ld	a,#38	;set octave
		add	a,(hl)
		out	(opl4_reg),a
		ld	a,e
		or %1000
		;set	3,a	;reverb?
		ex	af,af'
		opl4_wt
		ex	af,af'
		out	(opl4_data),a
		ld	a,#20	;set f-num
		add	a,(hl)
		ex	af,af'
		opl4_wt
		ex	af,af'
		out	(opl4_reg),a
		ld	a,d
		or 1		;bit 9 of wavenr
		ex	af,af'
		opl4_wt
		ex	af,af'
		out	(opl4_data),a
		ret

opl4_totallev:
		ld	e,(iy+q_mt_volume)
opl4_totallev0:
		ld	hl,mt_channel
		ld	a,#50	;set total level   
		add	a,(hl)
		ex	af,af'
		opl4_wt
		ex	af,af'
		out	(opl4_reg),a
		ld	d,0
		ld	hl,mt_vol_tab
		add	hl,de
		ld	a,(hl)
		rlca
		set	0,a	;level direct
		ex	af,af'
		opl4_wt
		ex	af,af'
		out	(opl4_data),a
		ld	hl,mt_channel
		ret

opl4_key:
		ld	hl,mt_channel
		opl4_wt
		ld	a,#68	;keyon
		add	a,(hl)
		out	(opl4_reg),a
		ld	e,(hl)
		ld	d,0
		ld	hl,mt_pansetting
		add	hl,de
		ld	a,(hl)
		ret

mt_settempous:
		ld	hl,(mt_tempo)
		ld	a,h	;tempo > 255  
		or	a
		jp	z,mt_settempous1
		ld	l,255	;dan r2=255  
mt_settempous1:
		opl4_wt
		ld	a,q_opl4_timer_reg
		out	(#c4),a
		opl4_wt
		ld	a,l
		neg
		out	(#c5),a

		opl4_wt
		ld	a,4	;R4
		out	(#c4),a
		opl4_wt
		ld	a,128	;RST
		out	(#c5),a
;:		ld	a,4
;:		out	(#c4),a
;:		opl4_wt
;:		ld	a,q_opl4_timer_on	;enable TIMER1
;:		out	(#c5),a
;:		opl4_wt

		ld	e,l
		ld	d,0
		or	a
		ld	hl,(mt_tempo)
		sbc	hl,de
		ld	(mt_tempo),hl
		ret

mt_command_0:
		xor	a
		ld	(mt_arpeggio),a
		jp	mt_return

mt_command_4:
		ld	l,(iy+q_mt_period)
		ld	h,(iy+q_mt_period+1)
		ld	(iy+q_mt_cmdtemp),h
		ld	(iy+q_mt_cmdtemp+1),l

		ld	a,(ix+3)
		and	%11110000
		jp	z,mt_command_41
		ld	b,a
		ld	a,(iy+q_mt_oldvibcmd)
		and	%00001111
		or	b
		ld	(iy+q_mt_oldvibcmd),a
mt_command_41:
		ld	a,(ix+3)
		and	%00001111
		jp	z,mt_return
		ld	b,a
		ld	a,(iy+q_mt_oldvibcmd)
		and	%11110000
		or	b
		ld	(iy+q_mt_oldvibcmd),a
		jp	mt_return

mt_command_7:
		ld	a,(ix+3)
		and	%11110000
		jp	z,mt_command_71
		ld	b,a
		ld	a,(iy+q_mt_oldtremcmd)
		and	%00001111
		or	b
		ld	(iy+q_mt_oldtremcmd),a
mt_command_71:
		ld	a,(ix+3)
		and	%00001111
		jp	z,mt_return
		ld	b,a
		ld	a,(iy+q_mt_oldtremcmd)
		and	%11110000
		or	b
		ld	(iy+q_mt_oldtremcmd),a
		jp	mt_return

mt_command_8:
		call	mt_command8
		jp	mt_return
;:		;Sets the current mt_channel's panning position.
;:		;Ranges from 00h (left) to FFh (right). 
;:		ld	a,(cmd8enable)
;:		or	a
;:		jp	z,mt_return
;:		ld	de,mt_pansetting
;:		ld	hl,(mt_channel)
;:		ld	h,0
;:		add	hl,de
;:		or	a
;:		ld	a,(ix+3)	;get panning value
;:		rrca
;:		rrca
;:		rrca
;:		rrca
;:		cp	9
;:		jp	c,mt_command_81
;:		dec	a
;mt_command_81
;:		add	a,9
;:		and	%1111
;:		ld	(hl),a	;set panning
;:		jp	mt_return

mt_command_9:
		;Sets the sample offset by starting another wave
		ld	a,(mt_cmd9enable)
		or	a
		jp	z,mt_return
		ld	a,(ix+3)
		or	a
		jp	z,mt_return
		ld	(iy+q_mt_sampnum),a
		jp	mt_return

mt_command_B:
		ld	a,(mt_song_pos)
		ld	b,a
		ld	a,(mt_song_length)
		dec	a
		cp	b
		jp	nz,mt_command_B1
		ld	a,(mt_repeat_song)
		or	a
		jp	nz,mt_command_B1
		ld	a,1
		ld	(mt_stopflag),a
		jp	mt_return
mt_command_B1:
		ld	a,(mt_vblanks+1)
		cp	1
		jp	nz,mt_return
		ld	a,(ix+3)
		;handle position queue
		ld	b,a
		ld	a,(mt_song_pos_queue)
		cp	255
		jp	z,mt_command_B2
		ld	b,a
mt_command_B2:
		ld	a,b

		ld	(mt_song_pos),a
		;reset the queue
		ld	a,255
		ld	(mt_song_pos_queue),a
		ld	a,128
		ld	(mt_song_step),a
		jp	mt_return

mt_command_C:
		ld	a,(ix+3)
		cp	#41
		jp	c,mt_command_C1
		ld	a,#40
mt_command_C1:
		ld	(iy+q_mt_volume),a
		call	opl4_totallev
		jp	mt_return

mt_command_D:
		ld	a,(mt_song_pos)
		ld	b,a
		ld	a,(mt_song_length)
		dec	a
		cp	b
		jp	nz,mt_command_D2
		ld	a,(mt_repeat_song)
		or	a
		jp	nz,mt_command_D2
		ld	a,1
		ld	(mt_stopflag),a
		jp	mt_return
mt_command_D2:
;:		ld	a,(mt_vblanks+1)
;:		cp	1
;:		jp	nz,mt_return
		ld	a,(mt_song_step)
		bit	7,a
		jp	nz,mt_return
		ld	a,(ix+3)
		ld	b,a
		and	%00001111
		ld	c,a
		ld	a,b
		and	%11110000
		rrca
		ld	b,a
		rrca
		rrca
		add	a,b
		add	a,c
		set	7,a
		ld	(mt_song_step),a
		ld	a,(mt_song_length)
		ld	b,a
		ld	a,(mt_song_pos)
		inc	a
		cp	b
		jp	c,mt_command_D1
		xor	a
mt_command_D1:
		;handle position queue
		ld	b,a
		ld	a,(mt_song_pos_queue)
		cp	255
		jp	z,mt_command_D3
		ld	b,a
mt_command_D3:
		ld	a,b
		ld	(mt_song_pos),a
		;reset the queue
		ld	a,255
		ld	(mt_song_pos_queue),a
		jp	mt_return

mt_command_E1:
		push	af
		ld	a,(ix+3)
		and	%1111
		ld	e,a
		ld	d,0
		ld	l,(iy+q_mt_period)
		ld	h,(iy+q_mt_period+1)
		scf
		ccf
		sbc	hl,de
		ld	de,108
		call	mt_rst20
		jp	nc,mt_command_E11
		ld	hl,108
mt_command_E11:
		ld	(iy+q_mt_period),l
		ld	(iy+q_mt_period+1),h
		call	opl4_fnum
		pop	af
		ret

mt_command_E2:
		push	af
		ld	a,(ix+3)
		and	%1111
		ld	e,a
		ld	d,0
		ld	l,(iy+q_mt_period)
		ld	h,(iy+q_mt_period+1)
		add	hl,de
		ld	de,907+1
		call	mt_rst20
		jp	c,mt_command_E21
		ld	hl,907
mt_command_E21:
		ld	(iy+q_mt_period),l
		ld	(iy+q_mt_period+1),h
		call	opl4_fnum
		pop	af
		ret

mt_command_E4:
		ld	a,(ix+3)
		and	%11
		ld	b,a
		ld	a,(iy+q_mt_wavectrl)
		and	%11111100
		or	b
		ld	(iy+q_mt_wavectrl),a
		xor	a
		ld	(iy+q_mt_vibtabpos),a
		jp	mt_return

mt_command_E5:
		ld	a,(ix+3)
		and	%1111
		ld	(mt_command_E5h),a
		jp	mt_return

mt_command_E7:
		ld	a,(ix+3)
		rlca
		rlca
		and	%1100
		ld	b,a
		ld	a,(iy+q_mt_wavectrl)
		and	%11110011
		or	b
		ld	(iy+q_mt_wavectrl),a
		xor	a
		ld	(iy+q_mt_tremtabpos),a
		jp	mt_return

mt_command_E8:
		;Sets the current mt_channel's panning position.
		;Ranges from 00h (left) to FFh (right). 
		ld	de,mt_pansetting
		ld	hl,(mt_channel)
		ld	h,0
		add	hl,de
		ld	a,(ix+3)	;get panning value
		and	%1111
		cp	8
		jp	nc,mt_command_E81
		inc	a
mt_command_E81:
		sub	8
		and	%1111
		ld	(hl),a	;set panning
		jp	mt_return

mt_command_E9:
		ld	a,(ix+3)
		and	%1111
		ld	(iy+q_mt_cmdtemp),a
		jp	mt_return

mt_command_EA:
		ld	a,(ix+3)
		rrca
		rrca
		rrca
		rrca
		and	%1111
		ld	b,a
		ld	a,(iy+q_mt_volume)
		add	a,b
		cp	#41
		jp	c,mt_command_EA1
		ld	a,#40
mt_command_EA1:
		ld	(iy+q_mt_volume),a
		call	opl4_totallev
		jp	mt_return

mt_command_EB:
		ld	a,(ix+3)
		and	%1111
		ld	b,a
		ld	a,(iy+q_mt_volume)
		sub	b
		jp	nc,mt_command_EA1
		xor	a
		jp	mt_command_EA1

mt_command_EC:
		ld	a,(ix+3)
		and	%1111
		jp	z,mt_command_EC1
		inc	a
		ld	(iy+q_mt_cmdtemp),a
		jp	mt_return
mt_command_EC1:
		xor	a
		ld	(iy+q_mt_volume),a
		jp	mt_return

mt_command_ED:
		ld	a,(ix+3)
		and	%1111
		jp	z,mt_return
		;inc	a	;ED0 makes no sense
		ld	(iy+q_mt_cmdtemp),a
		jp	mt_return_ED

mt_command_EE:
		ld	a,(ix+3)
		and	%1111
		ld	(mt_command_EEh),a
		jp	mt_return

mt_command_F:
		ld	a,(ix+3)
		cp	#20
		jp	nc,mt_command_F1
		or	a
		push	af
		call	z,mt_stop
		pop	af
		ld	(mt_vblanks+1),a
		ld	(mt_vblanks),a
		jp	mt_return
mt_command_F1:
		ld	e,(ix+3)
		ld	d,0
		ld	(mt_tempo+4),de	;store amiga settings
		
		if	q_opl4_timer_reg=3
		;calculate OPL4 timer2 setting
		;8*n2=2048-(62500/bmp)
		ld	hl,62500
		call	mt_div16
		;/8
		srl	b
		rr	c
		srl	b
		rr	c
		srl	b
		rr	c
		endif

		if	q_opl4_timer_reg=2
		;calculate OPL4 timer1 setting
		ld	hl,31250
		call	mt_div16
		endif

		;dec	bc		;round to higher
		ld	(mt_tempo+2),bc	;store OPL4 setting
		ld	(mt_tempo+0),bc
		call	mt_settempous		;activate the new setting
		jp	mt_return
